import numpy as np
import PIL
import pyx
import os
import nipype
from nipype.interfaces import fsl

import suspect

voxel_outline_color = pyx.color.rgb(1, 1, 0)
voxel_outline_width = pyx.style.linewidth(1.0)

beta_csf = 1
beta_gm = 0.78
beta_wm = 0.65

water_conc = 55556


def make_canvas(slice_array):
    scaling_factor = 255 / np.amax(slice_array)
    slice_image = PIL.Image.fromarray((slice_array * scaling_factor).astype(np.uint8))
    slice_canvas = pyx.canvas.canvas()
    slice_canvas.insert(pyx.bitmap.bitmap(0, 0, slice_image,
                                          height=slice_image.height),
                        [pyx.trafo.scale(1, -1, 0, slice_image.height / 2)])
    return slice_canvas


def get_voxel_path(volume, mrs):
    voxel_coords_mrs = [[-0.5, -0.5, -0.5],
                        [0.5, -0.5, -0.5],
                        [0.5, 0.5, -0.5],
                        [-0.5, 0.5, -0.5],
                        [-0.5, -0.5, -0.5],
                        [-0.5, -0.5, 0.5],
                        [0.5, -0.5, 0.5],
                        [0.5, 0.5, 0.5],
                        [-0.5, 0.5, 0.5],
                        [-0.5, -0.5, 0.5],
                        [0.5, -0.5, 0.5],
                        [0.5, -0.5, -0.5],
                        [0.5, 0.5, -0.5],
                        [0.5, 0.5, 0.5],
                        [-0.5, 0.5, 0.5],
                        [-0.5, 0.5, -0.5]]
    voxel_coords_volume = volume.from_scanner(mrs.to_scanner(voxel_coords_mrs))
    path_points = [pyx.path.moveto(*voxel_coords_volume[0, 0:2])] + \
                  [pyx.path.lineto(*point) for point in voxel_coords_volume[1:, 0:2]]
    path = pyx.path.path(*path_points)
    return path


def get_voxel_slices(volume, mrs):
    slices = [volume.resample(volume.coronal_vector,
                           volume.axial_vector,
                           shape=(1, 256, 256),
                           centre=(mrs.centre[0], volume.centre[1], volume.centre[2])),
              volume.resample(volume.sagittal_vector,
                              -volume.coronal_vector,
                              shape=(1, 256, 256),
                              centre=(volume.centre[0], volume.centre[1], mrs.centre[2])),
              volume.resample(volume.sagittal_vector,
                              volume.axial_vector,
                              shape=(1, 256, 256),
                              centre=(volume.centre[0], mrs.centre[1], volume.centre[2]))]

    scaling_factor = 255 / np.amax(volume)
    canvases = [make_canvas(slc * scaling_factor) for slc in slices]
    paths = [get_voxel_path(slc, mrs) for slc in slices]

    combined_canvas = pyx.canvas.canvas()
    for i in range(3):
        canvases[i].stroke(paths[i], [voxel_outline_color, voxel_outline_width])
        combined_canvas.insert(canvases[i], [pyx.trafo.scale(6 / 256).translated(6.5 * i, 0)])

    return combined_canvas


def _save_nifti(dicom_file, nifti_file=None):
    """
    This function is intended to be wrapped in a nipype Node and used to take a
    DICOM file (and other files from the same series in the same file), and
    save the volume in Nifti format instead. By default the output file will be
    in the folder created by nipype, which is usually temporary, but it can
    also be given an output path to save it somewhere specific.

    :param dicom_file:
    :param nifti_file:
    :return:
    """
    import suspect
    import os

    volume = suspect.image.load_dicom_volume(dicom_file)
    if nifti_file is None:
        nifti_file = os.path.join(os.getcwd(), "volume.nii")
    suspect.image.save_nifti(nifti_file, volume)
    return nifti_file


def classify_tissues(volume_path):
    ext = os.path.splitext(volume_path)[1]

    wf = nipype.Workflow(name="classify_wf")

    bet = nipype.Node(fsl.BET(frac=0.5,
                              robust=True),
                      name="bet")
    fast = nipype.Node(fsl.FAST(output_type="NIFTI",
                                number_classes=3),
                       name="fast")

    wf.connect([(bet, fast, [("out_file", "in_files")])])
    if ext.upper() in [".DCM", ".IMA"]:
        save_nifti = nipype.Node(nipype.Function(input_names=["dicom_file"],
                                                 output_names=["nifti_file"],
                                                 function=_save_nifti),
                                 name="save_nifti")
        save_nifti.inputs.dicom_file = volume_path
        wf.connect([(save_nifti, bet, [("nifti_file", "in_file")])])
    else:
        bet.inputs.in_file = os.path.abspath(volume_path)

    result = wf.run()

    for node in result.nodes():
        if node.name == "fast":
            wm = suspect.image.load_nifti(node.result.outputs.partial_volume_files[2])
            gm = suspect.image.load_nifti(node.result.outputs.partial_volume_files[1])
            csf = suspect.image.load_nifti(node.result.outputs.partial_volume_files[0])

    return wm, gm, csf


def segment_voxel(volume_path, mask):
    wm, gm, csf = classify_tissues(volume_path)
    voxel_volume = np.sum(mask)

    f_csf = float(np.sum(csf * mask) / voxel_volume)
    f_gm = float(np.sum(gm * mask) / voxel_volume)
    f_wm = float(np.sum(wm * mask) / voxel_volume)

    tissue_beta = f_csf * beta_csf + f_gm * beta_gm + f_wm * beta_wm
    corrected_water_conc = int(water_conc * tissue_beta)

    return {
        "wm": f_wm,
        "gm": f_gm,
        "csf": f_csf,
        "volume": voxel_volume,
        "tissue_beta": tissue_beta,
        "water_conc": corrected_water_conc,
        "water_att": gasparovic_attenuation_factor(f_wm, f_gm, f_csf)
    }


def gasparovic_attenuation_factor(f_wm, f_gm, f_csf, te=30, tr=2000, t1_m=1400, t2_m=200):
    """
    Calculates the attenuation factor that needs to be applied to obtain
    milliMolal absolute concentration units according to Gasparovich 2006
    which accounts for changes in the size of the water reference peak due to
    differences in T1 and T2 for grey and white matter and CSF. It also
    corrects for the fact that metabolites only appear in GM/WM and not in CSF.

    T1 and T2 values for water and metabolites are taken from Gussew 2012
    :param f_wm:
    :param f_gm:
    :param f_csf:
    :param te:
    :param tr:
    :param t1_m:
    :param t2_m:
    :return:
    """
    t1_wm = 1080
    t2_wm = 70
    t1_gm = 1820
    t2_gm = 100
    t1_csf = 4160
    t2_csf = 500

    r_wm = np.exp(-te / t2_wm) * (1 - np.exp(-tr / t1_wm))
    r_gm = np.exp(-te / t2_gm) * (1 - np.exp(-tr / t1_gm))
    r_csf = np.exp(-te / t2_csf) * (1 - np.exp(-tr / t1_csf))
    r_m = np.exp(-te / t2_m) * (1 - np.exp(-tr / t1_m))

    water_norm = f_wm * beta_wm + f_gm * beta_gm + f_csf * beta_csf
    fw_wm = beta_wm * f_wm / water_norm
    fw_gm = beta_gm * f_gm / water_norm
    fw_csf = beta_csf * f_csf / water_norm

    return (fw_wm * r_wm + fw_gm * r_gm + fw_csf * r_csf) / (1 - fw_csf) / r_m



import argparse
import numpy as np
import pyx
import suspect
import os

from cubric_mrs import voxel, table

# define some useful global parameters
voxel_outline_color = pyx.color.rgb(1, 1, 0)
voxel_outline_width = pyx.style.linewidth(1.0)
plot_linewidth = pyx.style.linewidth(0.01)
simple_plot_color = pyx.color.rgb(0, 0, 1)
sr_plot_color = pyx.color.rgb(1, 0, 0)
plot_linewidth_gaba = pyx.style.linewidth(0.03)
plot_color_gaba = pyx.color.rgb(0, 1, 0)
plot_color_residual = pyx.color.cmyk.Gray
plot_color_fit = pyx.color.rgb.red
plot_color_data = pyx.color.rgb.black
min_ppm = 0.2
max_ppm = 5.0


def analyse_press(data_path, t1_path=None, wref_path=None, out_path=None, out_csv=None, subject_id=None):
    data = suspect.io.load_twix(data_path)[:, 1]

    if subject_id is None:
        subject_id = data.metadata["patient_id"]

    num_channels = data.shape[-2]
    noise = np.moveaxis(data[:, :, -250:], 2, 0).reshape(num_channels, -1)

    white_data = suspect.processing.channel_combination.whiten(data, noise)
    if wref_path is not None:
        wref = suspect.io.load_twix(wref_path)[:, 1]
        white_wref = suspect.processing.channel_combination.whiten(wref, noise)
        channel_weights = suspect.processing.channel_combination.svd_weighting(np.mean(white_wref, axis=0))
        cc_wref = suspect.processing.channel_combination.combine_channels(white_wref, channel_weights)
        # assume no frequency drift over water signal, and doesn't matter for singlet peak anyway
        wref_final = np.mean(cc_wref, axis=0)
        # calculate the eddy current from the water signal
        ec = np.unwrap(np.angle(wref_final))
        # smooth the eddy current signal
        ec_smooth = suspect.processing.denoising.sliding_gaussian(ec, 32)
        ecc = np.exp(-1j * ec_smooth)
        wref_final *= ecc
        white_data *= ecc
    else:
        channel_weights = suspect.processing.channel_combination.svd_weighting(np.mean(white_data, axis=0))

    cc_data = suspect.processing.channel_combination.combine_channels(white_data, channel_weights)

    # re-bin into phase cycles and average over each cycle
    pc_data = np.mean(cc_data.reshape(cc_data.shape[0] // 16, 16, -1), axis=1)

    def correct_frequency_sr(target):
        def correct_fid(fid):
            frequency_shift, phase_shift = suspect.processing.frequency_correction.spectral_registration(fid, target)
            return fid.adjust_frequency(-frequency_shift).adjust_phase(-phase_shift)

        return correct_fid

    sr_data = np.mean(np.apply_along_axis(correct_frequency_sr(pc_data[0]), 1, pc_data), axis=0)

    one_page_canvas = pyx.canvas.canvas()
    files_dict = {
        "PRESS": os.path.basename(data_path)
    }
    if wref_path is not None:
        files_dict["WREF"] = os.path.basename(wref_path)
    if t1_path is not None:
        files_dict["T1"] = os.path.basename(t1_path)
    file_table = table.file_table(files_dict)
    one_page_canvas.insert(file_table,
                           [pyx.trafo.translate(1, 28)])

    if t1_path is not None:
        t1_path = os.path.abspath(t1_path)
        if os.path.splitext(t1_path)[1].upper() in [".IMA", ".DCM"]:
            t1 = suspect.image.load_dicom_volume(t1_path)
        elif os.path.splitext(t1_path)[1].upper() in [".NII"]:
            t1 = suspect.image.load_nifti(t1_path)
        elif t1_path.upper().endswith(".NII.GZ"):
            t1 = suspect.image.load_nifti(t1_path)
        else:
            print("could not load t1 from {}".format(os.path.splitext(t1_path)[1].upper()))
            exit(-1)
        voxel_canvases = voxel.get_voxel_slices(t1, data)
        segmentation = voxel.segment_voxel(t1_path, suspect.image.create_mask(data, t1))
        one_page_canvas.insert(voxel_canvases,
                               [pyx.trafo.translate(1, 28 - 0.5 - voxel_canvases.bbox().height())])

    water = wref_final if wref_path is not None else None
    tarquin_results = suspect.io.tarquin.process(sr_data, water, options={
        "w_conc": segmentation["water_conc"],
        "w_att": segmentation["water_att"]
    })

    tarquin_plot = plot_fitted_spectrum(tarquin_results["plots"]["data"],
                                        tarquin_results["plots"]["fit"] + tarquin_results["plots"]["baseline"])

    one_page_canvas.insert(tarquin_plot,
                           [pyx.trafo.translate(1, 28 - 1 - voxel_canvases.bbox().height() - tarquin_plot.height)])

    conc_table = table.metabolite_table(tarquin_results["metabolite_fits"],
                                        float(tarquin_results["metabolite_fits"]["TCr"]["concentration"]),
                                        "mM" if wref_path is not None else "A.U.")
    one_page_canvas.insert(conc_table,
                           [pyx.trafo.translate(11, 28 - 1 - voxel_canvases.bbox().height())])

    quality_table = table.quality_table(tarquin_results["quality"])
    #print(conc_table.bbox())
    one_page_canvas.insert(quality_table,
                           [pyx.trafo.translate(11, 28 - conc_table.bbox().height() - 1.5 - voxel_canvases.bbox().height())])

    properties_table = table.voxel_properties_table(segmentation)
    one_page_canvas.insert(properties_table,
                           [pyx.trafo.translate(1, 28 - 2 - voxel_canvases.bbox().height() - tarquin_plot.height)])

    output_sheet = pyx.document.page(one_page_canvas,
                                     paperformat=pyx.document.paperformat.A4,
                                     centered=False)
    output_document = pyx.document.document([output_sheet])

    output_document.writePDFfile(out_path)

    if out_csv is not None:
        output_csv(out_csv, subject_id, tarquin_results["metabolite_fits"])


def plot_fitted_spectrum(data, fit):
    data_range = np.amax(data.real) - np.amin(data.real)

    gaba_plot = pyx.graph.graphxy(width=9, height=6,
                                  x=pyx.graph.axis.linear(min=max_ppm, max=min_ppm),
                                  y=pyx.graph.axis.linear(
                                      min=np.amin(data.real) - 0.3 * data_range,
                                      max=np.amax(data.real) + 0.1 * data_range,
                                      parter=None
                                  ),
                                  key=pyx.graph.key.key(pos="tr")
                                  )

    gaba_plot.plot(pyx.graph.data.values(
        x=data.frequency_axis_ppm(),
        y=data.spectrum().real,
        title="Data"

    ),
        [pyx.graph.style.line([plot_linewidth, plot_color_data])]
    )

    gaba_plot.plot(pyx.graph.data.values(
        x=data.frequency_axis_ppm(),
        y=fit.real,
        title="Fit"
    ),
        [pyx.graph.style.line([plot_linewidth, plot_color_fit])]
    )

    gaba_plot.plot(pyx.graph.data.values(
        x=data.frequency_axis_ppm(),
        y=(data.real - fit.real - data_range * 0.15),
        title="Residual"
    ),
        [pyx.graph.style.line([plot_linewidth, plot_color_residual])]
    )

    return gaba_plot


def output_csv(filename, subject_id, concentrations):
    with open(filename, 'a') as fout:
        if fout.tell() == 0:
            # this is a new file, have to output the header line
            fout.write("SubjectID")
            for metabolite_name in sorted(concentrations.keys()):
                fout.write(", {0}, {0}_SD".format(metabolite_name))
        fout.write("\n")
        # now write the actual concentration values
        fout.write("{}".format(subject_id))
        for name, data in sorted(concentrations.items()):
            fout.write(", {}, {}".format(data["concentration"], data["sd"]))


def press_script():
    parser = argparse.ArgumentParser()

    parser.add_argument("--t1",
                        help="path to the T1 structural image in Nifti or DICOM format",
                        default=None)
    parser.add_argument("--press",
                        help="path to the press twix file",
                        required=True)
    parser.add_argument("--wref",
                        help="path to the water reference twix file",
                        default=None)
    parser.add_argument("--out_pdf",
                        help="path to save the one-page pdf output",
                        default="out.pdf")
    parser.add_argument("--out_csv",
                        help="path to save the concentration csv file",
                        default=None)
    parser.add_argument("--id",
                        help="override the patient id in output files",
                        default=None)

    args = parser.parse_args()

    analyse_press(args.press,
                  t1_path=args.t1,
                  wref_path=args.wref,
                  out_path=args.out_pdf,
                  out_csv=args.out_csv,
                  subject_id=args.id)



import pyx
import numpy as np


def file_table(files):
    table_runner = pyx.text.LatexRunner()
    table_runner.preamble("\\usepackage{tabularx}")

    file_strings = ["{} & : & {}".format(label, filename.replace("_", "\\_")) for label, filename in files.items()]
    table_string = """
\\begin{{tabularx}}{{19cm}}{{l c X}}
{}
\\end{{tabularx}}
""".format("\\\\\n".join(file_strings))
    return table_runner.text(0, 0, table_string, [pyx.text.valign.bottom])


def metabolite_table(concentrations, ref_cr, units="AU"):
    table_runner = pyx.text.LatexRunner()
    table_runner.preamble("\\usepackage{tabularx}")

    if ref_cr <= 0:
        ref_cr = np.inf

    metabolite_strings = ["{} & {} & {:.3f} & {}\\\\\n".format(
        name.replace("_", "\\_"), value["concentration"], float(value["concentration"]) / ref_cr, value["sd"]) for
        name, value in sorted(concentrations.items())]

    table_string = "\\begin{{tabularx}}{{9cm}}{{X r r r}}\nMetabolite & Conc/{} & /Cr & SD\%\\\\\n\hline\n{}\\end{{tabularx}}".format(
        units,
        "".join(metabolite_strings))

    table_latex = table_runner.text(0, 0, table_string, [pyx.text.valign.top])
    return table_latex


def quality_table(quality_info):
    table_runner = pyx.text.LatexRunner()
    table_runner.preamble("\\usepackage{tabularx}")
    # now display the fit quality metrics
    quality_strings = ["{} & {}\\\\\n".format(
        param, value) for param, value in sorted(quality_info.items())]

    quality_table_string = "\\begin{{tabularx}}{{9cm}}{{X r}}\nFit Parameters&\\\\\n\hline\n{}\\end{{tabularx}}".format(
        "".join(quality_strings))

    quality_table_latex = table_runner.text(0, 0, quality_table_string, [pyx.text.valign.top])
    return quality_table_latex


def voxel_properties_table(voxel_properties):
    table_runner = pyx.text.LatexRunner()
    table_runner.preamble("\\usepackage{tabularx}")

    property_string = """
    Voxel volume / mm$^3$ & {volume}\\\\
    Grey matter & {gm:.1%}\\%\\\\
    White matter & {wm:.1%}\\%\\\\
    CSF & {csf:.1%}\\%\\\\
    Water Conc. & {water_conc}\\\\
    Water Att. & {water_att:.3f}\\\\
    """.format(**voxel_properties).replace("%", "\%")

    property_table_string = "\\begin{{tabularx}}{{9cm}}{{X r}}\nVoxel Properties&\\\\\n\hline\n{}\\end{{tabularx}}".format(
        "".join(property_string))

    property_table_latex = table_runner.text(0, 0, property_table_string, [pyx.text.valign.top])
    return property_table_latex
